#!/usr/bin/env python
from __future__ import absolute_import
from __future__ import print_function
import os
if "PBS_O_WORKDIR" in os.environ:
    workingDir = os.environ["PBS_O_WORKDIR"]
    os.chdir(workingDir)

import sys
py_rumd_lib = {2:"../../../lib", 3:"../../../lib_py3"}[sys.version_info[0]]
sys.path.insert(1, py_rumd_lib)
sys.path.insert(1, "../../../Python")

import rumd
# it is essential that there is no doubt about which version is actually run!
print("File used to import the rumd module", rumd.__file__)

from rumd.Simulation import Simulation
from rumd.RUMD_Test import RUMD_Test

class Test_EMT_MgCu(RUMD_Test):
    """
    Binary metallic system system using EMT potential with NVT dynamics.
    """
    def __init__(self):
        RUMD_Test.__init__(self)
        self.numParticlesList = [1000]

    
    def RunSimulation(self, mode, numParticles, pb, tp):
        """Run one simulation in specified mode, with given N, pb, tp"""
        # set the number of time-steps depending on the mode
        self.numTimeSteps = {"short":          100,
                             "performance":100000,
                             "stress":   100000000}[mode]

        # sample and simulation
        startConfg = "startN%d.xyz.gz" % numParticles


        sim = Simulation(startConfg, pb, tp, verbose=False)

        if self.generate_reference_data:
            sim.SetOutputDirectory(self.ref_directory)
        else:
            sim.SetOutputDirectory(self.out_directory)


        # integrator
        itg = rumd.IntegratorNVT(timeStep=0.1, targetTemperature=0.8, thermostatRelaxationTime=10.0)


        potential = rumd.EMT_Potential()

        E0_Cu     = -3.510
        s0_Cu     = 1.413
        V0_Cu     = 1.994
        eta2_Cu   = 3.040
        kappa_Cu  = 4.944
        Lambda_Cu = 3.694
        n0_Cu     = 0.0637
    
        E0_Mg     = -1.487
        s0_Mg     = 1.766
        V0_Mg     = 2.230
        eta2_Mg   = 2.541
        kappa_Mg  = 4.435
        Lambda_Mg = 3.293
        n0_Mg     = 0.0355    
       
        potential.SetParams(0, E0_Mg, s0_Mg, V0_Mg, eta2_Mg, kappa_Mg, Lambda_Mg, n0_Mg)
        potential.SetParams(1, E0_Cu, s0_Cu, V0_Cu, eta2_Cu, kappa_Cu, Lambda_Cu, n0_Cu)
        sim.SetIntegrator(itg)
        sim.SetPotential(potential)

        sim.SetOutputMetaData("energies", thermostat_Ps=True)
        
        # tune output depending on the mode
        if mode == "short":
            sim.SetOutputScheduling("energies", "linear", interval=1)
            sim.SetOutputScheduling("trajectory", "linear", interval=10)
            sim.SetBlockSize(50)
        elif mode == "stress":
            sim.SetOutputScheduling("energies", "linear", interval=5000)
        else:
            assert mode == "performance"
            print("Note there is no output directory for performance" \
                  " test--the only output is the timing file")

        sim.Run(self.numTimeSteps,
                suppressAllOutput=(mode == "performance"))



if __name__ == "__main__":
    rumd_test = Test_EMT_MgCu()

    # standard short test to be called from makefile
    # For generating reference data and running stress and performance modes
    # there are corresponding scripts to be submitted to pbs

    sys.exit(rumd_test.RunShort("gen_ref" in sys.argv))

    # for individual runs, could do for example
    #sys.exit(rumd_test.Test("short", 1000, 16, 8))
