#!/usr/bin/env python
"""Checks that interactions between unlike types are correctly included also
when cutoff is longer than the like-like cutoffs"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import unittest
import os
import numpy
import sys
import tempfile

py_rumd_lib = {2:"../../../lib", 3:"../../../lib_py3"}[sys.version_info[0]]
sys.path.insert(1, py_rumd_lib)
sys.path.insert(1, "../../../Python")


import rumd
import rumd.Simulation
# it is essential that there is no doubt about which version is actually run!
print("File used to import the rumd module", rumd.__file__)


types = [0, 0, 1, 1]

pos = numpy.array([[0.000, 0.000, 0.000],
                   [1.500, 0.000, 0.000],
                   [0.000, 1.500, 0.000],
                   [0.000, 0.000, 2.500]])
N = len(types)

xyz = "%d\nioformat=2 " % N
xyz += "sim_box=RectangularSimulationBox,10.00,10.00,10.00 numTypes=2  "
xyz += "mass=1.000,1.000 columns=type,x,y,z\n"

for t, this_pos in zip(types, pos):
    xyz += "%d %.3f %.3f %.3f\n" % (t,
                                    this_pos[0],
                                    this_pos[1],
                                    this_pos[2])
    

def vLJ(sigma, epsilon, r):
    """ Evaluate the Lennard-Jones potential (no cutoff)"""
    return 4.0*epsilon*(pow(r/sigma, -12.0) - pow(r/sigma, -6.0))

class TestCutoffBug(unittest.TestCase):
    """Check that cutoff for unlike particles is correctly handled """
    def setUp(self):
        self.tdir = tempfile.mkdtemp()
        self.finp = self.tdir + '/test_cutoff_bug.xyz'
        fh = open(self.finp, 'w')
        fh.write(xyz)
        fh.close()

        self.sim = rumd.Simulation.Simulation(self.finp, verbose=False)


        # Create LJ pair-potential
        potential = rumd.Pot_LJ_12_6(cutoff_method=rumd.ShiftedPotential)
        # Set up parameters for Lennard-Jones interactions
        # i  j   Sig   Eps  Rcut (in units of Sigma_ij)

        self.cutoffs = {(0, 0): 2.0,
                        (1, 1): 2.0,
                        (0, 1): 3.0}

        self.vCut = {}

        for i, j in self.cutoffs.keys():
            potential.SetParams(i, j, 1.00, 1.00, self.cutoffs[(i, j)])
            self.vCut[(i, j)] = vLJ(1.0, 1.0, self.cutoffs[(i, j)])


        self.sim.SetPotential(potential)

    def test_cutoffs(self):
        """ Calculate potential energy by hand and compare to rumd CalcF"""
        pe = 0.0
        for idx in range(N):
            for jdx in range(idx+1, N):
                type_i = types[idx]
                type_j = types[jdx]
                if type_i <= type_j:
                    vc = self.vCut[(type_i, type_j)]
                else:
                    vc = self.vCut[(type_j, type_i)]
                r = numpy.sqrt(numpy.sum((pos[idx] - pos[jdx])**2))
                
                if r < self.cutoffs[(type_i, type_j)]:
                    pe += vLJ(1.00, 1.00, r) - vc
                
        self.sim.sample.CalcF()
        pe_rumd = self.sim.sample.GetPotentialEnergy()

        #print pe, pe_rumd
        self.assertLess(abs((pe_rumd-pe)), 1.e-6)

    def tearDown(self):
        import shutil
        if os.path.exists(self.finp):
            os.remove(self.finp)
        if os.path.exists(self.tdir):
            shutil.rmtree(self.tdir)


if __name__ == '__main__':
    unittest.main()
