#!/usr/bin/env python
"""Checks that interactions between unlike types are correctly included also
when cutoff is longer than the like-like cutoffs"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
import unittest
import os
import numpy
import sys
import tempfile

py_rumd_lib = {2:"../../../lib", 3:"../../../lib_py3"}[sys.version_info[0]]
sys.path.insert(1, py_rumd_lib)
sys.path.insert(1, "../../../Python")


import rumd
import rumd.Simulation
# it is essential that there is no doubt about which version is actually run!
print("File used to import the rumd module", rumd.__file__)

types = [0, 0, 1, 1]

pos = numpy.array([[1.200, 0.000, 0.000],
                   [0.000, 1.400, 0.000],
                   [0.000, 0.000, 1.600],
                   [0.000, 0.000, 0.000]])
N = len(types)

xyz = "%d\nioformat=2 " % N
xyz += "sim_box=RectangularSimulationBox,10.00,10.00,10.00 numTypes=2  "
xyz += "mass=1.000,1.000 columns=type,x,y,z\n"

for t, this_pos in zip(types, pos):
    xyz += "%d %.3f %.3f %.3f\n" % (t,
                                    this_pos[0],
                                    this_pos[1],
                                    this_pos[2])


def vLJ(sigma, epsilon, r):
    """ Evaluate the Lennard-Jones potential (no cutoff)"""
    return 4.0*epsilon*(pow(r/sigma, -12.0) - pow(r/sigma, -6.0))

class TestExclusionList(unittest.TestCase):
    """Check that the exclusion list and exclusion by type
    work as they should """
    def setUp(self):
        self.tdir = tempfile.mkdtemp()
        self.finp = self.tdir + '/test_exclusion_list.xyz'
        fh = open(self.finp, 'w')
        fh.write(xyz)
        fh.close()


    def calculate_potential(self, exclude_pair=None,
                            exclude_type=None, sort=False):
        """ calculate potential energy for the test configuration with a certain
            pair excluded, or a certain type excluded, using either n2 or sort
            method for neighbor-list"""
        sim = rumd.Simulation.Simulation(self.finp, verbose=False)
        # Create LJ pair-potential
        potential = rumd.Pot_LJ_12_6(cutoff_method=rumd.NoShift)
        # Set up parameters for Lennard-Jones interactions
        # i  j   Sig   Eps  Rcut (in units of Sigma_ij)

        for i, j in [(0, 0), (0, 1), (1, 1)]:
            potential.SetParams(i, j, 1.00, 1.00, 5.0)

        sim.SetPotential(potential)

        if exclude_pair is not None:
            ex_i, ex_j = exclude_pair
            potential.SetExclusion(ex_i, ex_j)
            potential.CopyExclusionListToDevice()
        if exclude_type is not None:
            ex_pair_i, ex_pair_j = exclude_type
            print("excluding type %d-type %d" % (ex_pair_i, ex_pair_j))
            potential.SetExclusionType(ex_pair_i, ex_pair_j)

        if sort:
            sim.sample.SetSortingScheme(rumd.SORT_X)
            sim.sample.SortParticles()

        sim.sample.CalcF()
        pe_rumd = sim.sample.GetPotentialEnergy()

        pe = 0.0
        for idx in range(N):
            for jdx in range(idx+1, N):
                if exclude_pair is not None and idx == ex_i and jdx == ex_j:
                    print("excluding pair", ex_i, ex_j)
                    continue
                type_i = types[idx]
                type_j = types[jdx]
                if exclude_type is not None and ((type_i == ex_pair_i and
                                                    type_j == ex_pair_j) or
                                                    (type_i == ex_pair_j and
                                                    type_j == ex_pair_i)):
                    continue

                r = numpy.sqrt(numpy.sum((pos[idx] - pos[jdx])**2))
                pe += vLJ(1.00, 1.00, r)


        return pe, pe_rumd

    def test_calc_energy(self):
        """ Calculate potential energy by hand and compare to rumd CalcF, just
        to check basic operation before doing exclusions """

        pe_full, rumd_pe_full = self.calculate_potential()
        #print pe_full, rumd_pe_full
        self.assertLess(abs((rumd_pe_full - pe_full)), 1.e-6)

    def test_exclusion_list(self):
        """ Calculate potential energy by hand and compare to rumd CalcF,
        excluding pair 0,1"""

        pe_ex_pair, rumd_pe_ex_pair = self.calculate_potential(exclude_pair=
                                                               (0, 1))
        self.assertLess(abs((rumd_pe_ex_pair - pe_ex_pair)), 1.e-6)

    def test_exclusion_list_sort(self):
        """ Calculate potential energy by hand and compare to rumd CalcF,
        excluding pair 0,1 and sorting on x values"""

        pe_ex_pair, rumd_pe_ex_pair = self.calculate_potential(exclude_pair=
                                                               (0, 1),
                                                               sort=True)
        self.assertLess(abs((rumd_pe_ex_pair - pe_ex_pair)), 1.e-6)

    def test_exclusion_type(self):
        """ Calculate potential energy by hand and compare to rumd CalcF,
        excluding interactions between type 0 particles"""
        pe_ex_type, rumd_pe_ex_type = self.calculate_potential(exclude_type=
                                                               (0, 0))
        self.assertLess(abs((rumd_pe_ex_type - pe_ex_type)), 1.e-6)


    def tearDown(self):
        import shutil
        if os.path.exists(self.finp):
            os.remove(self.finp)
        if os.path.exists(self.tdir):
            shutil.rmtree(self.tdir)


if __name__ == '__main__':
    unittest.main()
